/* 
 * File:   descriptors_impl2.h
 * Author: tabrown
 *
 * Created on June 30, 2016, 11:53 AM
 */

#ifndef DESCRIPTORS_IMPL2_H
#define	DESCRIPTORS_IMPL2_H

typedef intptr_t tagptr_t;
typedef intptr_t seqbits_t;

#include <string.h>

#define DESC_INIT_ALL(descArray, macro_seqBitsNew, numProcesses) { \
    for (int i=0;i<(numProcesses);++i) { \
        (descArray)[i].seqBits = macro_seqBitsNew(0); \
    } \
}

///**
// * pointer to descriptor (listed least to most significant):
// * 1-bit: is descriptor
// * 10-bits: thread id
// * remaining bits: sequence number (for avoiding the ABA problem)
// */
//typedef intptr_t tagptr_t;
//typedef intptr_t seqbits_t;

/**
 * seqbits_t corresponds to the seqBits field of the descriptor.
 * it contains the mutable fields of the descriptor and a sequence number.
 * the width, offset and mask for the sequence number is defined below.
 * this sequence number width, offset and mask are also shared by tagptr_t.
 *
 * in particular, for any tagptr_t x and seqbits_t y, the sequence numbers
 * in x and y are equal iff x&MASK_SEQ == y&MASK_SEQ (despite the differing
 * types of x and y).
 * 
 * tagptr_t consists of a triple <seq, tid, testbit>.
 * these three fields are defined by the TAGPTR_ macros below.
 */

#ifndef WIDTH_SEQ
    #define WIDTH_SEQ 48
#endif
#define OFFSET_SEQ 14
#define MASK_SEQ ((uintptr_t)((1LL<<WIDTH_SEQ)-1)<<OFFSET_SEQ) /* cast to avoid signed bit shifting */
#define UNPACK_SEQ(tagptrOrSeqbits) (((uintptr_t)(tagptrOrSeqbits))>>OFFSET_SEQ)

#define TAGPTR_OFFSET_USER 0
#define TAGPTR_OFFSET_TID 3
#define TAGPTR_MASK_USER ((1<<TAGPTR_OFFSET_TID)-1) /* assumes TID is next field after USER */
#define TAGPTR_MASK_TID (((1<<OFFSET_SEQ)-1)&(~(TAGPTR_MASK_USER)))
#define TAGPTR_UNPACK_TID(tagptr) ((int) ((((tagptr_t) (tagptr))&TAGPTR_MASK_TID)>>TAGPTR_OFFSET_TID))
#define TAGPTR_UNPACK_PTR(descArray, tagptr) (&(descArray)[TAGPTR_UNPACK_TID((tagptr))])
#define TAGPTR_NEW(tid, seqBits, userBits) ((tagptr_t) (((UNPACK_SEQ(seqBits))<<OFFSET_SEQ) | ((tid)<<TAGPTR_OFFSET_TID) | (tagptr_t) (userBits)<<TAGPTR_OFFSET_USER))
// assert: there is no thread with tid DUMMY_TID that ever calls TAGPTR_NEW
#define LAST_TID (TAGPTR_MASK_TID>>TAGPTR_OFFSET_TID)
#define TAGPTR_STATIC_DESC(id) ((tagptr_t) TAGPTR_NEW(LAST_TID-1-id, 0))
#define TAGPTR_DUMMY_DESC(id) ((tagptr_t) TAGPTR_NEW(LAST_TID, id<<OFFSET_SEQ))

#define comma ,

#define SEQBITS_UNPACK_FIELD(seqBits, mask, offset) \
    ((((seqbits_t) (seqBits))&(mask))>>(offset))
// TODO: make more efficient version "SEQBITS_CAS_BIT"
// TODO: change sequence # unpacking to masking for quick comparison
// note: if there is only one subfield besides seq#, then the third if-block is redundant, and you should just return false if the cas fails, since the only way the cas fails and the field being cas'd contains still old is if the sequence number has changed.
#define SEQBITS_CAS_FIELD(successBit, fldSeqBits, snapSeqBits, oldval, val, mask, offset) { \
    seqbits_t __v = (fldSeqBits); \
    while (1) { \
        if (UNPACK_SEQ(__v) != UNPACK_SEQ((snapSeqBits))) { \
            (successBit) = false; \
            break; \
        } \
        if ((successBit) = __sync_bool_compare_and_swap(&(fldSeqBits), \
                (__v & ~(mask)) | (oldval), \
                (__v & ~(mask)) | ((val)<<(offset)))) { \
            break; \
        } \
        __v = (fldSeqBits); \
        if (SEQBITS_UNPACK_FIELD(__v, (mask), (offset)) != (oldval)) { \
            (successBit) = false; \
            break; \
        } \
    } \
}
// TODO: change sequence # unpacking to masking for quick comparison
// note: SEQBITS_FAA_FIELD would be very similar to SEQBITS_CAS_FIELD; i think one would simply delete the last if block and change the new val from (val)<<offset to (val&mask)+1.
#define SEQBITS_WRITE_FIELD(fldSeqBits, snapSeqBits, val, mask, offset) { \
    seqbits_t __v = (fldSeqBits); \
    while (UNPACK_SEQ(__v) == UNPACK_SEQ((snapSeqBits)) \
            && SEQBITS_UNPACK_FIELD(__v, (mask), (offset)) != (val) \
            && !__sync_bool_compare_and_swap(&(fldSeqBits), __v, \
                    (__v & ~(mask)) | ((val)<<(offset)))) { \
        __v = (fldSeqBits); \
    } \
}
#define SEQBITS_WRITE_BIT(fldSeqBits, snapSeqBits, mask) { \
    seqbits_t __v = (fldSeqBits); \
    while (UNPACK_SEQ(__v) == UNPACK_SEQ((snapSeqBits)) \
            && !(__v&(mask)) \
            && !__sync_bool_compare_and_swap(&(fldSeqBits), __v, (__v|(mask)))) { \
        __v = (fldSeqBits); \
    } \
}

// WARNING: uses a GCC extension "({ })". to get rid of this, use an inline function.
#define DESC_SNAPSHOT(descType, descArray, descDest, tagptr, sz) ({ \
    descType *__src = TAGPTR_UNPACK_PTR((descArray), (tagptr)); \
    memcpy((descDest), __src, (sz)); \
    __asm__ __volatile__ ("":::"memory"); /* prevent compiler from reordering read of __src->seqBits before (at least the reading portion of) the memcpy */ \
    (UNPACK_SEQ(__src->seqBits) == UNPACK_SEQ((tagptr))); \
})
#define DESC_READ_FIELD(successBit, fldSeqBits, tagptr, mask, offset) ({ \
    seqbits_t __seqBits = (fldSeqBits); \
    successBit = (__seqBits & MASK_SEQ) == ((tagptr) & MASK_SEQ); \
    SEQBITS_UNPACK_FIELD(__seqBits, (mask), (offset)); \
})
#define DESC_NEW(descArray, macro_seqBitsNew, tid) &(descArray)[(tid)]; { /* note: only the process invoking this following macro can change the sequence# */ \
    seqbits_t __v = (descArray)[(tid)].seqBits; \
    (descArray)[(tid)].seqBits = macro_seqBitsNew(__v); \
    /*__sync_synchronize();*/ \
}
#define DESC_INITIALIZED(descArray, tid) \
    (descArray)[(tid)].seqBits += (1<<OFFSET_SEQ);

#endif	/* DESCRIPTORS_IMPL2_H */

